"""
////////////////////////////////////////////////////////////////////////////////////////////
//                  analyze_sweep.py
// Log sine sweep analyzer (partner to generate_sweep.py)

// Author: Mark Tillotson
// Updated: 2020-01-17
// Language: Python3
// Version: 1.0
// Update History:
// Requirements:  numpy, scipy, PySoundFile, matplotlib

// v1.0 - Initial version 2020-01-17
// Getting started:

python3 analyze_sweep.py input_file.wav
-- defaults to 96000 SPS, 10 to 24kHz, 20 second sweep

// optional arguments:

--fs, --minf, --maxf, --duration

python3 analyze_sweep.py input_file.wav --fs 44100 --minf 20 --maxf 18000 --duration 10

note that the arguments must match those used to generate the source sweep file

Also there's a --distort flag for self-testing, produces even harmonics and also 3rd harmonic.

////////////////////////////////////////////////////////////////////////////////////////////
"""


import optparse
from math import sqrt, log10, pi, cos

import numpy as np
from numpy.fft import fft

from scipy.signal import convolve, chirp, resample, resample_poly

import matplotlib.pyplot as plt

from soundfile import SoundFile


# the parameters fs, minf, maxf and duration must match those used in generating the sweep

parser = optparse.OptionParser ()
parser.add_option ('--fs', action = 'store', type = float, default = 96000)
parser.add_option ('--minf', action = 'store', type = float, default = 10)
parser.add_option ('--maxf', action = 'store', type = float, default = 24000)
parser.add_option ('--duration', action = 'store', type = float, default = 20.0)
# This is for self-testing, adds in some distortion if set
parser.add_option ('--distort', action = 'store_true', default = False)
options, args = parser.parse_args ()

assert len(args) == 1, "Need single filename argument"


FS = options.fs

N = int (options.duration * FS)


# ramp up first 1/24th and ramp down last 1/24th of chirp with Tukey window
TUKEY_SAMPLES = N // 24

def tukey (v, size):
    if len(v) < 2*size:
        raise ValueError ("Tukey window size too big for array")
    tuk = 0.5 * (1.0 - np.cos (pi * np.arange(size) / size))
    v[0:size] *= tuk
    v[-size:] *= tuk[::-1]


# The deconvolution inverse sweep
def create_inverted_swept():
    r = pow (options.minf / options.maxf, np.arange (N) / N)
    return r * chirp (np.arange (0, N/FS, 1/FS),
                      options.maxf,
                      N / FS, options.minf,
                      method = 'logarithmic',
                      phi = -90)


# the size of the FFT used for each harmonic
EXCERPT_LEN = 4096

COLOURS = ['b', 'r', 'g', 'b', 'c', 'y', 'm', 'r', 'g', 'b', 'c', 'y', 'm']

# Can only get to about 80% of sweep max frequency before artifacts appear
MAX_DISPLAY_FREQ = int (0.80 * options.maxf / FS * EXCERPT_LEN)


def read_wavfile (filename):
    sndfile = SoundFile (filename)
    rate = sndfile.samplerate
    channels = sndfile.channels
    data = sndfile.read (always_2d = True)
    print ("SoundFile read, shape %s, rate %s, format %s, subtype %s" % (data.shape, rate, sndfile.format, sndfile.subtype))
    bits = 16 if sndfile.subtype == 'PCM_16' else (24 if sndfile.subtype == 'PCM_24' else None)
    print ("bit depth: %s" % bits)
    return data, rate, bits


# for self-testing
def apply_distortion (v):
    distorted = 0.001 * abs(v) + v + 0.01 * v*v*v
    return distorted - np.sum (distorted) / len(distorted)


# for showing the sweeps for sanity-check that the files match
def display_waveforms (v, inv_v):
    fig = plt.figure ()
    ax = fig.add_subplot(211)
    ax.set_title ("wave")
    ax.plot (np.arange (0, N, 50)/FS, v[0 : N : 50], 'k')
    ax.grid()
    ax = fig.add_subplot(212)
    ax.set_title ("invert")
    ax.plot (np.arange (0, N, 50)/FS, inv_v[0 : N : 50], 'k')
    ax.grid()
    plt.show()


if __name__ == '__main__':
    filename = args[0]
    v, rate, bits = read_wavfile (filename)
    FS = rate
    assert FS == options.fs, "sample rate mismatch"

    # pick out last channel if more than one
    if len(v.shape) == 2:
        v = v[:, v.shape[1]-1]
    assert len(v) >= N

    # if distort test mode, apply some obvious distortion
    if options.distort:
        v = apply_distortion (v)

    # find halfway point up the initial ramp (assumes flat freq response, or at least a low-pass characteristic, note)
    half_val = np.max (abs (v)) / 2
    for i in range (len (v)):
        if abs(v[i]) >= half_val:
            ramp_middle = i
            break
    ramp_start = ramp_middle - TUKEY_SAMPLES//2

    # now pick out the chirp out from the whole vector as we know where it starts
    v = v[ramp_start : ramp_start + N]

    # get the inverse convolution of the chirp's impulse response
    inv_v = create_inverted_swept()
    tukey (inv_v, TUKEY_SAMPLES)

    display_waveforms (v, inv_v)

    # convolve to get the impulse response
    convolution = convolve (inv_v, v, mode='same', method='auto') / N


    # search for the highest point, to compensate for any small temporal offset
    top_val = np.max (abs (convolution))
    for i in range (len (convolution)):
        if abs (convolution[i]) == top_val:
            zero_point = i
            break


    harmonic_delays = [ int (N * (log10(harm) / log10 (options.maxf / options.minf))) for harm in range (1, 10) ]

    # check we have enough samples to get whole spectrum for each harmonic
    for i in range (1, 10-1):
        assert harmonic_delays[i] - harmonic_delays[i-1] > EXCERPT_LEN


    fig = plt.figure ()
    ax = fig.add_subplot(111)

    # accumulate harmonics into this in the power domain
    total_distortion = np.ones ([MAX_DISPLAY_FREQ], dtype = float) / 1e30
    
    for harm in range (1, 10):
        offset = - harmonic_delays [harm-1]

        excerpt = convolution [zero_point + offset - EXCERPT_LEN//2 : zero_point + offset + EXCERPT_LEN//2]

        spectrum = abs (fft (excerpt[::-1], EXCERPT_LEN) [0 : MAX_DISPLAY_FREQ])
        
        if harm == 1:
            # the harmonic plots are relative to the fundamental response
            reference_spectrum = spectrum
            # get value at 1kHz
            reference_point = reference_spectrum [int (1000 / FS * EXCERPT_LEN)]
            
            # plot fundamental absolute response, with 1kHz defined as 0dB
            ax.semilogx (FS / EXCERPT_LEN * np.arange (MAX_DISPLAY_FREQ) / harm,
                         20 * (np.log10 (spectrum) - log10 (reference_point)),
                         COLOURS [harm-1])
        else:
            # for harmonics we have to scale the spectra down to the match the fundamental
            new_size = int (round (MAX_DISPLAY_FREQ / harm))
            # power is relative to fundamental
            power = spectrum / reference_spectrum
            power = power * power

            # uses resample_poly() to scale smoothly, and resample() to correct for out-by-one issues
            total_distortion [0 : new_size] += abs (resample (resample_poly (power, 1, harm), new_size))
            # all harmonics are displayed relative to fundamental, ie dBc
            ax.semilogx (FS / EXCERPT_LEN * np.arange (MAX_DISPLAY_FREQ) / harm,
                         20 * (np.log10 (spectrum) - np.log10 (reference_spectrum)),
                         COLOURS [harm-1])

        
    # total distortion curve
    ax.semilogx (FS / EXCERPT_LEN * np.arange (MAX_DISPLAY_FREQ),
                 10 * np.log10 (total_distortion), # - 20 * np.log10 (reference_spectrum),
                 'k')


    ax.legend (['fundamental'] + [ 'H%s' % h for h in range (2, 10) ] + ['total'], loc = 'upper left')
    ax.set_xlabel ('Hz')
    ax.set_ylabel ('dB')
    ax.set_ylim (-170, 10)
    ax.set_xlim (10, 22000)
    ax.set_yticks (range (-170, 10, 10))
    ax.set_xticks ([10, 20, 30, 40, 50, 60, 70, 80, 90,
                    100, 200, 300, 400, 500, 600, 700, 800, 900,
                    1000, 2000, 3000, 4000, 5000, 6000, 7000, 8000, 9000,
                    10000, 20000])
    ax.grid()
    plt.show ()

    

"""
////////////////////////////////////////////////////////////////////////////////////////////
//                                TERMS OF USE: MIT License
////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this 
// software and associated documentation files (the "Software"), to deal in the Software 
// without restriction, including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////////////////
"""
