"""
////////////////////////////////////////////////////////////////////////////////////////////
//                  generate_sweep.py
// Log sine sweep generator output wav file (and other formats supported by PySoundFile)

// Author: Mark Tillotson
// Updated: 2020-01-17
// Language: Python3
// Version: 1.0
// Update History:
// Requirements:  numpy, scipy, PySoundFile

// v1.0 - Initial version 2020-01-17

// Getting started:

python3 generate_sweep.py output_file.wav
-- defaults to 96000 SPS, 10 to 24kHz, 20 second sweep

// optional arguments:

--fs, --minf, --maxf, --duration

python3 generate_sweep.py output_file.wav --fs 44100 --minf 20 --maxf 18000 --duration 10


////////////////////////////////////////////////////////////////////////////////////////////
"""


import optparse
from math import pi

import numpy as np

from scipy.signal import chirp

import soundfile

parser = optparse.OptionParser ()
parser.add_option ('--fs', action = 'store', type = int, default = 96000)
parser.add_option ('--minf', action = 'store', type = float, default = 10)
parser.add_option ('--maxf', action = 'store', type = float, default = 24000)
parser.add_option ('--duration', action = 'store', type = float, default = 20.0)
parser.add_option ('--bits', action = 'store', type = int, default = 24)
parser.add_option ('--mono', action = 'store_true', default = False)
options, args = parser.parse_args ()

assert len(args) == 1, "Takes just a filename"
assert options.bits == 16 or options.bits == 24


filename = args[0]

FS = int (options.fs)

N = int (FS * options.duration)

TUKEY_SAMPLES = N // 24

def tukey (v, size):
    if len(v) < 2*size:
        raise ValueError ("Tukey window size too big for array")
    tuk = 0.5 * (1.0 - np.cos (pi * np.arange(size) / size))
    v[0:size] *= tuk
    v[-size:] *= tuk[::-1]



def create_swept ():
    return chirp (np.arange (0, N/FS, 1/FS),
                  options.minf,
                  N/FS,
                  options.maxf,
                  method = 'logarithmic',
                  phi = -90)

    

if __name__ == '__main__':
    
    v = create_swept()  # generate raw chirp
    
    tukey(v, TUKEY_SAMPLES)  # ramp at start and end
    
    samples = np.array ((0x40000000) * v, dtype = 'int32')  # convert to int32

    # pad with zeros front and back
    t = np.zeros ([int (N * 1.5)], 'int32')
    t [ N//4 : N + N//4] = samples
    samples = t

    # duplicate for stereo
    if not options.mono:
        t = np.zeros ([len(samples), 2], dtype = 'int32')
        t [:, 0] = samples
        t [:, 1] = samples
        samples = t
        print (samples.shape)

    # write it
    soundfile.write (filename, samples, FS, subtype = ('PCM_24' if options.bits == 24 else 'PCM_16'))

    

"""
////////////////////////////////////////////////////////////////////////////////////////////
//                                TERMS OF USE: MIT License
////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this 
// software and associated documentation files (the "Software"), to deal in the Software 
// without restriction, including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////////////////
"""
